-- ZIP and ZIP SFX files
-- (c) FSC, 2007

LOCAL_FILE_HEADER_SIGNATURE     =   0x04034b50
CENTRAL_FILE_HEADER_SIGNATURE   =   0x02014b50
END_OF_CENTRAL_DIR_SIGNATURE    =   0x06054b50
DATA_DESCRIPTOR_SIGNATURE       =   0x08074b50
ARCHIVE_EXTRA_DATA_SIGNATURE    =   0x08064b50
DIGITAL_SIGNATURE               =   0x05054b50

local MonthTable = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" }

--
-- number and store all files starting from 1
--
local LocalFile = 1
local CentralFile = 1

local LocalFiles = { }
local CentralFiles = { }

--
-- show all defined files
--
function ShowFiles()
    if LocalFile > 1 then
        CrLf()
        local Str = ""
        for i = 1, LocalFile - 1 do
            Str = Str .. "\r" .. LocalFiles[i]
        end
        BlockComment("\rThe following files are defined in LOCAL_FILE_HEADER blocks:\r" .. Str .. "\r")
    end
    if CentralFile > 1 then
        CrLf()
        local Str = ""
        for i = 1, CentralFile - 1 do
            Str = Str .. "\r" .. CentralFiles[i]
        end
        BlockComment("\rThe following files are defined in CENTRAL_FILE_HEADER blocks:\r" .. Str .. "\r")
    end
end

--
-- skip everything (that is, show as unknown data) until we find local file header
--
local ZipPos = FindMasked32(LOCAL_FILE_HEADER_SIGNATURE)

if not ZipPos then
    BlockComment("\rNo ZIP archives found in the current file\r\r", true)
    return
end

if ZipPos ~= 0 then
    MaxInlineElements = 16
    Hex()
    BYTE("Data", "probably SFX code", ZipPos)
    CrLf()
end

--
-- get binary representation of given value (up to 32 bits)
-- Warning: just for clearness we use non-standard binary representation with "b" suffix (like in Assembler, not in C)
--
local function ToBinary(val)
    local NumBits = val.Bits
    if NumBits > 32 then error("PrintBinary: too many bits") end
    local Mask = 2 ^ (NumBits - 1)
    local Val = ToUI32(val)
    local Str = ""
    for i = 1, NumBits do
        if TestBit(Val, Mask) then Str = Str .. "1" else Str = Str .. "0" end
        Mask = Mask / 2
    end
    return Str .. "b"
end

--
-- Extra field
--
local function FormatFileTime()
    local x = PeekFileTime()
    if not x then return "invalid FILETIME structure" end
    return string.format("%d %s %d %d:%02d:%02d", x.Day, MonthTable[x.Month], x.Year, x.Hour, x.Minute, x.Second)
end

local function FormatUnixTime()
    local x = PeekUnixTime()
    if not x then return "invalid Unix time" end
    return string.format("%d %s %d %d:%02d:%02d", x.Day, MonthTable[x.Month], x.Year, x.Hour, x.Minute, x.Second)
end

local function _ExtraField000a()
    TypeWidth = 8
    NameWidth = 17
    Hex()
    WORD("Tag", "0x000a: PKWARE Win95/WinNT FileTimes")
    Dec()
    WORD("TSize", "Total Data Size for this block (should be 32)")
    
    CrLf()
    Hex()
    LONG("Unknown1")
    LONG("Unknown2")
    
    CrLf()
    Hex()
    DWORD64("ModificationTime", FormatFileTime())
    DWORD64("AccessTime", FormatFileTime())
    DWORD64("CreationTime", FormatFileTime())
end

local function _ExtraField5455()
    NameWidth = 8
    Hex()
    WORD("Tag", "0x5455: extended filestamp")
    Dec()
    local TSize = PeekUI16()
    WORD("TSize", "Total Data Size for this block")
    local Start = GetFilePosition()
    
    if TSize > 0 then CrLf() end
    Hex()
    if TSize > BytesProcessed(Start) then BYTE("Flags", "info bits") end
    if TSize > BytesProcessed(Start) then DWORD("ModTime", FormatUnixTime() .. " time of last modification (UTC/GMT)") end
    if TSize > BytesProcessed(Start) then DWORD("AcTime", FormatUnixTime() .. " time of last access (UTC/GMT)") end
    if TSize > BytesProcessed(Start) then DWORD("CrTime", FormatUnixTime() .. " time of original creation (UTC/GMT)") end
    
    if TSize > BytesProcessed(Start) then BYTE("UnknownData", nil, TSize - BytesProcessed(Start)) end
end

local function _ExtraField7855()
    NameWidth = 8
    Hex()
    WORD("Tag", "0x7855: Info-ZIP Unix")
    Dec()
    local TSize = PeekUI16()
    WORD("TSize", "total data size for this block")
    local Start = GetFilePosition()
    
    if TSize > 0 then CrLf() end
    Hex()
    if TSize > BytesProcessed(Start) then DWORD("AcTime", FormatUnixTime() .. " time of last access (UTC/GMT)") end
    if TSize > BytesProcessed(Start) then DWORD("ModTime", FormatUnixTime() .. " time of last modification (UTC/GMT)") end
    if TSize > BytesProcessed(Start) then DWORD("UID", "Unix user ID") end
    if TSize > BytesProcessed(Start) then DWORD("GID", "Unix group ID") end

    if TSize > BytesProcessed(Start) then BYTE("UnknownData", nil, TSize - BytesProcessed(Start)) end
end

ExtraField000a = struct("FILE_TIMES", _ExtraField000a)
ExtraField5455 = struct("EXTENDED_TIMESTAMP", _ExtraField5455)
ExtraField7855 = struct("INFO_ZIP_UNIX", _ExtraField7855)

local function _ExtraField(Len)
    while Len > 0 do
        local Start = GetFilePosition()
        local Tag = PeekUI16()
        if     Tag == 0x000a then ExtraField000a("FileTimes")
        elseif Tag == 0x5455 then ExtraField5455("ExtFilestamp")
        elseif Tag == 0x7855 then ExtraField7855("UnixExtraField")
        else BYTE("Unknown", nil, Len)
        end
        Len = Len - BytesProcessed(Start)
        if Len > 0 then CrLf() end
    end
end

ExtraField = struct("EXTRA_FIELD", _ExtraField)

--
-- Data descriptor
--
local function _DataDescriptor()
    NameWidth = 17
    if PeekUI32() ~= DATA_DESCRIPTOR_SIGNATURE then error("DATA_DESCRIPTOR_SIGNATURE expected") end
    Hex()
    DWORD("Signature", "must be 0x08074b50")
    CrLf()
    DWORD("CRC32", "CRC")
    HexDec()
    DWORD("CompressedSize", "size of compressed file")
    DWORD("UncompressedSize", "size of uncompressed file")
end

DataDescriptor = struct("DATA_DESCRIPTOR", _DataDescriptor)

--
-- decode MS-DOS Date and Time fields
--
local function MSDOSTime()
    local Sec2 = ToUI8(ReadBits(5))
    local Min = ToUI8(ReadBits(6))
    local Hour = ToUI8(ReadBits(5))
    SkipBytes(-2)
    return string.format("%d:%02d:%02d", Hour, Min, Sec2 * 2)
end

local function MSDOSDate()
    local Day = ToUI8(ReadBits(5))
    local Month = ToUI8(ReadBits(4))
    local MonthStr = MonthTable[Month] or "???"
    local Year1980 = ToUI16(ReadBits(7))
    SkipBytes(-2)
    return string.format("%d %s %d", Day, MonthStr, Year1980 + 1980)
end

--
-- Version fields
--
local HOSTSYSTEM  =
{
    [0] =   "FAT file system (DOS, OS/2, NT) + PKZIPW 2.50 VFAT, NTFS",     -- 0
            "Amiga",                                                        -- 1
            "VMS (VAX or Alpha AXP)",                                       -- 2
            "Unix",                                                         -- 3
            "VM/CMS",                                                       -- 4
            "Atari",                                                        -- 5
            "HPFS file system (OS/2, NT 3.x)",                              -- 6
            "Macintosh",                                                    -- 7
            "Z-System",                                                     -- 8
            "CP/M",                                                         -- 9
            "TOPS-20 [supposedly PKZIPW 2.50 NTFS]",                        -- 10
            "NTFS file system (NT) [used by Info-ZIP, only]",               -- 11
            "SMS/QDOS",                                                     -- 12
            "Acorn RISC OS",                                                -- 13
            "VFAT file system (Win95, NT) [Info-ZIP reservation, unused]",  -- 14
            "MVS",                                                          -- 15
            "BeOS (BeBox or PowerMac)",                                     -- 16
            "Tandem"                                                        -- 17
}

function _Version()
    TypeWidth = 5
    NameWidth = 11
    Dec()
    local Version = PeekUI8()
    BYTE("Version", math.floor(Version / 10) .. "." .. math.floor(Version % 10))

    local HostComment = HOSTSYSTEM[PeekUI8()] or "unused"
    BYTE("HostSystem", HostComment)
end

Version = struct("VERSION", _Version)

--
-- local file header flags
--
local function _LFHFlags()
    TypeWidth = 5
    NameWidth = 20

    -- we need to know the compression method used in order to interpret flags
    SkipBytes(2)
    local Meth = PeekUI16()
    SkipBytes(-2)

    Dec()
    BITFIELD("WORD", "Encrypted", "file is encrypted if this is not 0", 1, ToUI8)
    
    if     Meth == 6 then
        BITFIELD("WORD", "DictionarySize", "0/1 - 4K/8K sliding dictionary was used", 1, ToUI8)
        BITFIELD("WORD", "NumberOfTrees", "0/1 - 3/2 Shannon-Fano trees were used to encode thesliding dictionary output", 1, ToUI8)

    elseif Meth == 8 then
        local Meth8 = ToUI8(PeekBits(2))
        local Meth8Comment = ({ [0] =   "Normal (-en) compression option was used.",
                                        "Maximum (-ex) compression option was used.",
                                        "Fast (-ef) compression option was used.",
                                        "Super Fast (-es) compression option was used." })[Meth8]
        
        BITFIELD("WORD", "CompressionOption", Meth8Comment, 2, ToBinary)
    else
        BITFIELD("WORD", "Undefined", "unused for anything but compression methods 6 & 8", 2, ToBinary)
    end
    
    local NoCRC32 = ToUI8(PeekBits(1))
    BITFIELD("WORD", "DataDescFollows", "1 means that the fields crc-32, compressed size and uncompressed size are set to zero in the local header and data descriptor follows the data", 1, ToUI8)
    BITFIELD("WORD", "Reserved", "unused", 1, ToUI8)
    BITFIELD("WORD", "Cpd", "1 means that the file is compressed patched data", 1, ToUI8)
    
    if ToUI16(PeekBits(10)) ~= 0 then
        BITFIELD("WORD", "Reserved", "unused", 10, ToBinary)
    else
        SkipBits(10)
    end
    return NoCRC32 == 1
end

LFHFlags = struct("FLAGS", _LFHFlags)

--
-- show compression type
--
local function Compression()
    local MethodComment = ({ [0] =  "The file is stored (no compression)",
                                    "The file is Shrunk",
                                    "The file is Reduced with compression factor 1",
                                    "The file is Reduced with compression factor 2",
                                    "The file is Reduced with compression factor 3",
                                    "The file is Reduced with compression factor 4",
                                    "The file is Imploded",
                                    "Reserved for Tokenizing compression algorithm",
                                    "The file is Deflated",
                                    "Reserved for enhanced Deflating",
                                    "PKWARE Data Compression Library Implodingend"
                           })[PeekUI16()] or "unknown"
    Dec()
    WORD("CompressionMethod", MethodComment)
end

--
-- Local file header
--
function _LocalFileHeader()
    TypeWidth = 6
    NameWidth = 18
    
    if PeekUI32() ~= LOCAL_FILE_HEADER_SIGNATURE then error("LOCAL_FILE_HEADER_SIGNATURE expected") end
    
    Hex()
    DWORD("Signature", "local file header signature (0x04034b50)")

    CrLf()
    Version("Version", "version needed to extract")
    
    CrLf()
    local DataDescPresent = LFHFlags("Flags")
    
    CrLf()
    Compression()
    
    CrLf()
    Hex()
    WORD("LastModFileTime", MSDOSTime())
    WORD("LastModFileDate", MSDOSDate())
    
    CrLf()
    DWORD("CRC32", "CRC")
    HexDec()
    local CompSize = PeekUI32()
    DWORD("CompressedSize", "size of compressed file")
    DWORD("UncompressedSize", "size of uncompressed file")
    
    CrLf()
    Dec()
    local NameLen = PeekUI16()
    WORD("FileNameLength", "length of the filename")
    local EFLen = PeekUI16()
    WORD("ExtraFieldLength", "length of the extra field")

    -- show and remember file name
    if NameLen > 0 then
        LocalFiles[LocalFile] = PeekBytes(NameLen).Data
        char__("FileName", "name of the file (without terminating zero)", NameLen, true)
    else
        LocalFiles[LocalFile] = ""
    end
    Hex()
    
    if EFLen > 0 then
        CrLf()
        ExtraField("ExtraField", nil, EFLen)
    end
    
    CrLf()
    if DataDescPresent then
        -- we don't know how much packed data we have and must search for Data Descriptor
        local DDPos = FindMasked32(DATA_DESCRIPTOR_SIGNATURE)
        if not DDPos then error("Could not find DATA_DESCRIPTOR_SIGNATURE") end
        CompSize = DDPos - GetFilePosition().Offset
    end
    
    if CompSize == 0 then BlockComment("\rSize of compressed data is 0, that is, there is no compressed data for the file\r\r") end
    BYTE("CompData", nil, CompSize)
    
    if DataDescPresent then
        CrLf()
        DataDescriptor("DataDesc")
    end

    LocalFile = LocalFile + 1
end

LocalFileHeader = struct("LOCAL_FILE_HEADER", _LocalFileHeader)

--
-- File header (part of central directory structure)
--
local function _CentralFileHeader()
    TypeWidth = 6
    NameWidth = 18

    if PeekUI32() ~= CENTRAL_FILE_HEADER_SIGNATURE then error("CENTRAL_FILE_HEADER_SIGNATURE expected") end
    
    Hex()
    DWORD("Signature", "central file header signature (0x02014b50)")

    CrLf()
    Version("VersionMade", "version made by")
    CrLf()
    Version("VersionExtract", "version needed to extract")

    CrLf()
    local DataDescPresent = LFHFlags("Flags")
    
    CrLf()
    Compression()

    CrLf()
    Hex()
    WORD("LastModFileTime", MSDOSTime())
    WORD("LastModFileDate", MSDOSDate())

    CrLf()
    DWORD("CRC32", "CRC")
    HexDec()
    local CompSize = PeekUI32()
    DWORD("CompressedSize", "size of compressed file")
    DWORD("UncompressedSize", "size of uncompressed file")

    CrLf()
    Dec()
    local NameLen = PeekUI16()
    WORD("FileNameLength", "length of the filename")
    local EFLen = PeekUI16()
    WORD("ExtraFieldLength", "length of the extra field")
    local FCLen = PeekUI16()
    WORD("FileCommentLength", "length of comment fields")

    CrLf()
    Dec()
    WORD("DiskNumberStart", "the number of the disk on which this file begins")

    CrLf()
    Hex()
    WORD("InternalFileAttr", "bit0 set to 1 means that the file is an ASCII or text file. All other bits are unused")
    
    DWORD("ExternalFileAttr", "external file attributes, host-system dependent. For MS-DOS, the low order byte is the MS-DOS directory attribute byte")
    
    CrLf()
    DWORD("LHOffset", "relative offset of local header")
    
    CrLf()
    if NameLen > 0 then
        CentralFiles[CentralFile] = PeekBytes(NameLen).Data
        char__("FileName", "name of the file (without terminating zero)", NameLen, true)
    else
        CentralFiles[CentralFile] = ""
    end
    Hex()
    
    if EFLen > 0 then
        CrLf()
        ExtraField("ExtraField", nil, EFLen)
    end

    if FCLen > 0 then char__("FileComment", "file comment", FCLen, true) end

    CentralFile = CentralFile + 1
end

CentralFileHeader = struct("CENTRAL_FILE_HEADER", _CentralFileHeader)

--
-- End of central dir
--
local function _EndCentralDir()
    TypeWidth = 6
    NameWidth = 17
    if PeekUI32() ~= END_OF_CENTRAL_DIR_SIGNATURE then error("END_OF_CENTRAL_DIR_SIGNATURE expected") end

    Hex()
    DWORD("Signature", "end of central dir record signature (0x06054b50)")

    CrLf()
    Dec()
    WORD("ThisDisk", "number of this disk")
    WORD("CDirStartDisk", "number of the disk with the start of the central directory")
    WORD("ThisDiskEntries", "total number of entries in the central dir on this disk")
    WORD("TotalEntries", "total number of entries in the central dir")
    
    CrLf()
    HexDec()
    DWORD("CDirSize", "size of the central directory")
    DWORD("CDirOffset", "offset of start of central directory with respect to the starting disk number")
    
    CrLf()
    Dec()
    local ZFCommentLen = PeekUI16()
    WORD("ZipCommentLength", "zipfile comment length")
    if ZFCommentLen > 0 then char__("ZipComment", "zipfile comment", ZFCommentLen, true) end
end

EndCentralDir = struct("END_OF_CENTRAL_DIR", _EndCentralDir)

--
-- Archive extra data record
--
local function _ArchiveExtraData()
    TypeWidth = 6
    NameWidth = 17
    if PeekUI32() ~= ARCHIVE_EXTRA_DATA_SIGNATURE then error("ARCHIVE_EXTRA_DATA_SIGNATURE expected") end

    Hex()
    DWORD("Signature", "archive extra data signature (0x08064b50)")
    local Len = PeekUI32()
    DWORD("ExtraFieldLength", "extra field length")
    if Len > 0 then BYTE("ExtraField", "extra field data", Len) end
end

ArchiveExtraData = struct("ARCHIVE_EXTRA_DATA_RECORD", _ArchiveExtraData)

--
-- digital signature
--
local function _DigitalSignature()
    TypeWidth = 6
    NameWidth = 11
    if PeekUI32() ~= DIGITAL_SIGNATURE then error("DIGITAL_SIGNATURE expected") end
    
    Hex()
    DWORD("Signature", "header signature (0x05054b50)")
    local Len = PeekUI16()
    WORD("DataLength", "size of data")
    if Len > 0 then BYTE("Data", "signature data", Len) end
end

DigitalSignature = struct("DIGITAL_SIGNATURE", _DigitalSignature)

--
-- the whole Zip file
--
local function _ZipFile()
    while true do
        local Tag = PeekUI32()
        if     Tag == LOCAL_FILE_HEADER_SIGNATURE   then LocalFileHeader("LocalFile" .. LocalFile)
        elseif Tag == CENTRAL_FILE_HEADER_SIGNATURE then CentralFileHeader("CentralFile" .. CentralFile)
        elseif Tag == END_OF_CENTRAL_DIR_SIGNATURE  then EndCentralDir("EndOfCentralDir") break
        elseif Tag == ARCHIVE_EXTRA_DATA_SIGNATURE  then ArchiveExtraData("ArchiveExtraData")
        elseif Tag == DIGITAL_SIGNATURE             then DigitalSignature("Signature")
        else error(string.format("Unknown Zip signature: 0x%08x", Tag))
        end
        CrLf()
    end
end

ZipFile = struct("ZIPFILE", _ZipFile)
